#!/usr/bin/python

import asynchat
import asyncore
import socket
import sys
import string
import os
import fcntl
import ldap
import ConfigParser
import syslog


class policy_server(asyncore.dispatcher):

    def __init__(self, conf):
        asyncore.dispatcher.__init__(self)
	self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
	self.set_reuse_addr()
	self.bind((conf['host'], string.atoi(conf['port'])))
	self.listen(10)
	self.uuid = 0
	self.conf = conf

    def handle_accept(self):
        self.uuid = self.uuid + 1
        policy_receiver(self, self.accept(), self.uuid, self.conf)

class policy_receiver(asynchat.async_chat):

    def __init__(self, server, (conn, addr), uuid, conf):
        asynchat.async_chat.__init__(self, conn)
        self.set_terminator('\n')
        self.server = server
        self.id = uuid
	self.conf = conf
	if conf['debug'] == "TRUE":
            self.debug = True
        else:
            self.debug = False
        self.buffer = ''
	self.policy_input = {}
	if self.debug:
            syslog.syslog('Connection from: '+addr[0]+':'+str(addr[1]))

    def collect_incoming_data(self, data):
        self.buffer = self.buffer + data

    def found_terminator(self):
        data = self.buffer
        self.buffer = ''
        if self.debug:
            syslog.syslog(syslog.LOG_DEBUG, '<== (%d) %s' % (self.id, repr(data)))
        if (data != ""):
            sep_pos = data.find("=")
            if (sep_pos == 0):
                self.close()
                return
            self.policy_input[data[:sep_pos]] = data[(sep_pos+1):]
        else:
            ret = self.check_policy(self.policy_input)
            self.policy_input = {}
            if self.debug:
                syslog.syslog(syslog.LOG_DEBUG, '==> (%d) action=%s' % (self.id, str(ret)))
            self.push('action='+str(ret)+'\n\n')

    def handle_close(self):
        self.close()

    def check_policy(self, pol):
        sender = pol["sender"].lower()
	sender_dom = sender[sender.find("@"):] # @sender.domain.tld
        recipient = pol["recipient"].lower()
	recipient_dom = recipient[recipient.find("@"):] # @recipient.domain.tld

        sender_check_filter = self.conf['sender_check_filter'].replace('$s', sender).replace('$S', sender_dom).replace('$r', recipient).replace('$R', recipient_dom)
        recipient_check_filter = self.conf['recipient_check_filter'].replace('$s', sender).replace('$S', sender_dom).replace('$r', recipient).replace('$R', recipient_dom)
	base_dn = self.conf['base_dn']

        if self.debug:
            syslog.syslog(syslog.LOG_DEBUG, '--- (%d) sender_check_filter: %s' % (self.id, sender_check_filter))
            syslog.syslog(syslog.LOG_DEBUG, '--- (%d) recipient_check_filter: %s' % (self.id, recipient_check_filter))
            syslog.syslog(syslog.LOG_DEBUG, '--- (%d) base_dn: %s' % (self.id, base_dn))

        try:
            l = ldap.initialize(self.conf['url'])
            l.simple_bind_s(self.conf['bind_dn'], self.conf['bind_pwd'])
        except ldap.LDAPError, e:
            syslog.syslog(syslog.LOG_ERR, '==> (%d) ### LDAP Exception:[%s]' % (self.id, str(e)))
            return "DEFER_IF_PERMIT (xpd) Service temporarily unavailable"

        ret = "DUNNO"

        check = "SENDER_DUNNO"
	obj = {}
        userobj = {}
        domobj = {}
        # start checking the sender is allowed by the recipient user or domain
        try:
            r = l.search_s(base_dn, ldap.SCOPE_SUBTREE, sender_check_filter)
        except ldap.LDAPError, e:
            syslog.syslog(syslog.LOG_ERR, '==> (%d) ### LDAP Exception:[%s]' % (self.id, str(e)))
            return "DEFER_IF_PERMIT (xpd) Service temporarily unavailable"

        if len(r) > 0:
            #lowecase keys
            for i in r[0][1].keys():
                obj[i.lower()] = r[0][1][i]
            for i in obj["objectclass"]:
                if i.lower() == self.conf['user_objectclass']:
                    userobj = obj
                elif i.lower() == self.conf['domain_objectclass']:
                    domobj = obj
            if len(r) > 1:
                #lowecase keys
                for i in r[1][1].keys():
                    obj[i.lower()] = r[1][1][i]
                for i in obj["objectclass"]:
                    if i.lower() == self.conf['user_objectclass']:
                        userobj = obj
                    elif i.lower() == self.conf['domain_objectclass']:
                        domobj = obj

            if self.debug:
                if len(userobj) == 0:
                    syslog.syslog(syslog.LOG_ERR, 'sender_check user entry not found, filter error ??')
                if len(domobj) == 0:
                    syslog.syslog(syslog.LOG_ERR, 'sender_check domain entry not found, filter error ??')
                try:
                    for i in userobj[self.conf["permit_sender"]]:
                        syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- user permitted sender: %s" % (self.id, i))
                except KeyError:
                    syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- user permitted sender not found" % (self.id))
                try:
                    for i in userobj[self.conf["deny_sender"]]:
                        syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- user denied sender: %s" % (self.id, i))
                except KeyError:
                    syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- user denied sender not found" % (self.id))
                try:
                    for i in domobj[self.conf["permit_sender"]]:
                        syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- domain permitted sender: %s" % (self.id, i))
                except KeyError:
                    syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- domain permitted sender not found" % (self.id))
                try:
                    for i in domobj[self.conf["deny_sender"]]:
                        syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- domain denied sender: %s" % (self.id, i))
                except KeyError:
                    syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- domain denied sender not found" % (self.id))

            #check if this sender is permitted by the user entry
            try:
                for i in userobj[self.conf["permit_sender"]]:
                    if i == sender:
                        check = "SENDER_OK"
    		    break
                    if i == sender_dom:
                        check = "SENDER_DOM_OK"
    		    break
            except KeyError:
                pass

            if check != "SENDER_OK":
                try:
                    for i in userobj[self.conf["deny_sender"]]:
                        if i == sender:
                            check = "SENDER_REJECT"
                        if i == sender_dom and check != "SENDER_DOM_OK":
                            check = "SENDER_REJECT"
                        if i == '@' and check != "SENDER_DOM_OK":
                            check = "SENDER_REJECT"
                except KeyError:
                    pass

            if check == "SENDER_DUNNO":
                #check if this sender is permitted by the domain entry
                try:
                    for i in domobj[self.conf["permit_sender"]]:
                        if i == sender:
                            check = "SENDER_OK"
                        if i == sender_dom:
                            check = "SENDER_DOM_OK"
                except KeyError:
                    pass

                if check != "SENDER_OK":
                    try:
                        for i in domobj[self.conf["deny_sender"]]:
                            if i == sender:
                                check = "SENDER_REJECT"
                            if i == sender_dom and check != "SENDER_DOM_OK":
                                check = "SENDER_REJECT"
                            if i == '@' and check != "SENDER_DOM_OK":
                                check = "SENDER_REJECT"
                    except KeyError:
                        pass

            if check == "SENDER_OK" or check == "SENDER_DOM_OK":
                ret = self.conf['ok']
            elif check == "SENDER_REJECT":
                ret = "REJECT Sender Blacklisted"

        check = "RECIPIENT_DUNNO"
	obj = {}
        userobj = {}
        domobj = {}

        # start checking the recipient is allowed by the sender user or domain
        try:
            r = l.search_s(base_dn, ldap.SCOPE_SUBTREE, recipient_check_filter)
        except ldap.LDAPError, e:
            syslog.syslog(syslog.LOG_ERR, '==> (%d) ### LDAP Exception:[%s]' % (self.din, str(e)))
            return "DEFER_IF_PERMIT (xpd) Service temporarily unavailable"

        if len(r) > 0:
            #lowercase keys
            for i in r[0][1].keys():
                obj[i.lower()] = r[0][1][i]
            for i in obj["objectclass"]:
                if i.lower() == self.conf['user_objectclass']:
                    userobj = obj
                elif i.lower() == self.conf['domain_objectclass']:
                    domobj = obj
            if len(r) > 1:
                for i in r[1][1].keys():
                    obj[i.lower()] = r[1][1][i]
                for i in obj["objectclass"]:
                    if i.lower() == self.conf['user_objectclass']:
                        userobj = obj
                    elif i.lower() == self.conf['domain_objectclass']:
                        domobj = obj
            if userobj == False:
                syslog.syslog(syslog.LOG_ERR, 'recipient_check user entry not found, filter error ??')
            if domobj == False:
                syslog.syslog(syslog.LOG_ERR, 'recipient_check domain entry not found, filter error ??')

            if self.debug:
                try:
                    for i in userobj[self.conf["permit_recipient"]]:
                        syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- user permitted recipient: %s" % (self.id, i))
                except KeyError:
                    syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- user permitted recipient not found" % (self.id))
                try:
                    for i in userobj[self.conf["deny_recipient"]]:
                        syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- user denied recipient: %s" % (self.id, i))
                except KeyError:
                    syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- user denied recipient not found" % (self.id))
                try:
                    for i in domobj[self.conf["permit_recipient"]]:
                        syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- domain permitted recipient: %s" % (self.id, i))
                except KeyError:
                    syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- domain permitted recipient not found" % (self.id))
                try:
                    for i in domobj[self.conf["deny_recipient"]]:
                        syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- domain denied recipient: %s" % (self.id, i))
                except KeyError:
                    syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --- domain denied recipient not found" % (self.id))

            #check if this recipient is permitted by the user entry
            try:
                for i in userobj[self.conf["permit_recipient"]]:
                    if i == recipient:
                        check = "RECIPIENT_OK"
                    if i == recipient_dom:
                        check = "RECIPIENT_DOM_OK"
            except KeyError:
                pass

            if check != "RECIPIENT_OK":
                try:
                    for i in userobj[self.conf["deny_recipient"]]:
                        if i == recipient:
                            check = "RECIPIENT_REJECT"
                        if i == recipient_dom and check != "RECIPIENT_DOM_OK":
                            check = "RECIPIENT_REJECT"
                        if i == '@' and check != "RECIPIENT_DOM_OK":
                            check = "RECIPIENT_REJECT"
                except KeyError:
                    pass

            if check == "RECIPIENT_DUNNO":
                #check if this recipient is permitted by the domain entry
                try:
                    for i in domobj[self.conf["permit_recipient"]]:
                        if i == recipient:
                            check = "RECIPIENT_OK"
                        if i == recipient_dom:
                            check = "RECIPIENT_DOM_OK"
                except KeyError:
                    pass

                if check != "RECIPIENT_OK":
                    try:
                        for i in domobj[self.conf["deny_recipient"]]:
                            if i == recipient:
                                check = "RECIPIENT_REJECT"
                            if i == recipient_dom and check != "RECIPIENT_DOM_OK":
                                check = "RECIPIENT_REJECT"
                            if i == '@' and check != "RECIPIENT_DOM_OK":
                                check = "RECIPIENT_REJECT"
                    except KeyError:
                        pass

            if check == "RECIPIENT_OK" or check == "RECIPIENT_DOM_OK":
                ret = self.conf['ok']
            elif check == "RECIPIENT_REJECT":
                ret = "REJECT Recipient Blacklisted"

        try:
            l.unbind()
        except ldap.LDAPError, e:
            syslog.syslog(syslog.LOG_ERR, '==> (%d) ### LDAP Exception:[%s]' % (self.id, str(e)))
            return "DEFER_IF_PERMIT (xpd) Service temporarily unavailable"

        if self.debug:
            syslog.syslog(syslog.LOG_DEBUG, "--- (%d) --> %s" % (self.id, ret))
        return ret


if __name__ == '__main__':

    def become_daemon():
        try:
            pid = os.fork()
            if pid > 0:
                #parent exits
		sys.exit(0)
        except OSError, e:
            syslog.syslog(syslog.LOG_ERR, "fork failed, can't become daemon: %d (%s)" % (e.errno, e.strerror))
            print >>sys.stderr, "fork failed, can't become daemon: %d (%s)" % (e.errno, e.strerror)
            sys.exit(1)

        os.chdir("/")
	os.setsid()
	os.umask(0)

    #import pdb
    #pdb.set_trace()
 
    if len(sys.argv) < 3:
	print 'Usage: %s -c <config_file>' % sys.argv[0]
	sys.exit(1)
 
    syslog.openlog('xsec-policyd')
    config = ConfigParser.SafeConfigParser()
    config.read(sys.argv[2])

    try:
        conf = {}
        conf['host'] = config.get('basic', 'host')
        conf['port'] = config.get('basic', 'port')
        conf['url'] = config.get('ldap', 'url')
        conf['base_dn'] = config.get('ldap', 'base_dn')
        conf['sender_check_filter'] = config.get('ldap', 'sender_check_filter')
        conf['recipient_check_filter'] = config.get('ldap', 'recipient_check_filter')
        conf['user_objectclass'] = config.get('ldap', 'user_objectclass').lower()
        conf['domain_objectclass'] = config.get('ldap', 'domain_objectclass').lower()
        conf['permit_sender'] = config.get('ldap', 'permit_sender').lower()
        conf['permit_recipient'] = config.get('ldap', 'permit_recipient').lower()
        conf['deny_sender'] = config.get('ldap', 'deny_sender').lower()
        conf['deny_recipient'] = config.get('ldap', 'deny_recipient').lower()
    except ConfigParser.NoOptionError:
        syslog.syslog(syslog.LOG_ERR, 'Missing mandatory parameters, bailing out!')
        print >>sys.stderr, "Missing mandatory parameters in config file, bailing out!"
        sys.exit(1)

    try:
        conf['debug'] = config.get('basic', 'debug').upper()
    except ConfigParser.NoOptionError:
        conf['debug'] = "FALSE"

    try:
        conf['ok'] = config.get('basic', 'ok_command')
    except ConfigParser.NoOptionError:
        conf['ok'] = "DUNNO"

    try:
        conf['bind_dn'] = config.get('ldap', 'bind_dn')
    except ConfigParser.NoOptionError:
        conf['bind_dn'] = ""

    try:
        conf['bind_pwd'] = config.get('ldap', 'bind_pwd')
    except ConfigParser.NoOptionError:
        conf['bind_pwd'] = ""

    try:
        if config.get('basic', 'daemon') == 'TRUE':
            become_daemon()
            syslog.syslog(syslog.LOG_INFO, 'Run as daemon')
            try:
                pid_file = config.get('basic', 'pid_file')
                pid = os.getpid()
                f = open(pid_file, "w")
                f.write(str(pid)+"\n")
                f.close()
            except ConfigParser.NoOptionError:
                syslog.syslog(syslog.LOG_INFO, 'no pid file in configuration file')

    except ConfigParser.NoOptionError:
        syslog.syslog(syslog.LOG_INFO, 'NOT run as daemon')

    ps = policy_server(conf)
    asyncore.loop()
